############################################################
# Helper definitions

# Requires and Conflicts interact with the Provides;
# therefore if rpmbuild was expected to add a disttag to the Provides
# but is not able to, skip the tests with Requires and Conflicts.
# Obsoletes doesn't interact with the Provides; it must honor the real NEVRD.
#
# There are versions of rpmbuild that add a DistTag as a tag,
# but not to the Provides. So the distinction makes sense.
define overlaps
$(foreach p,$(2),\
	$(if $(and $(SKIP_RPMBUILD_PROV_DISTTAG),$($(p)_disttag)),,\
		$(call installable_seq,$(p),req$(1),$(3)) \
		$(call noninstallable,$(p),confl$(1),$(3))) \
	$(call obsoleted,$(p),obsol$(1),$(3)))
endef

define overlaps_but_not_obsoletes
$(foreach p,$(2),\
	$(if $(and $(SKIP_RPMBUILD_PROV_DISTTAG),$($(p)_disttag)),,\
		$(call installable_seq,$(p),req$(1),$(3)) \
		$(call noninstallable,$(p),confl$(1),$(3))) \
	$(call nonobsoleted,$(p),obsol$(1),$(3)))
endef

define not_overlaps
$(foreach p,$(2),\
	$(if $(and $(SKIP_RPMBUILD_PROV_DISTTAG),$($(p)_disttag)),,\
		$(call noninstallable_seq,$(p),req$(1),$(3)) \
		$(call installable,$(p),confl$(1),$(3))) \
	$(call nonobsoleted,$(p),obsol$(1),$(3)))
endef

# The constraint is Requires or Conflicts (or Obsoletes).
define mk_constraint_rpms
confl$(1)_preamble_etc = Conflicts: $$($(1))
$(call mk_bin_rpm,confl$(1))

req$(1)_preamble_etc = Requires: $$($(1))
$(call mk_bin_rpm,req$(1))

obsol$(1)_preamble_etc = Obsoletes: $$($(1))
$(call mk_bin_rpm,obsol$(1))
endef

############################################################
# Test cases

# Constraints with an epoch.
GreaterEpoch = dummy > 0:$(dummy_version)-$(dummy_release)
$(eval $(call mk_constraint_rpms,GreaterEpoch))
LessEpoch = dummy < 0:$(dummy_version)-$(dummy_release)
$(eval $(call mk_constraint_rpms,conflLessEpoch))
define EqualEpoch
dummy = 0:$(dummy_version)-$(dummy_release)
endef
$(eval $(call mk_constraint_rpms,EqualEpoch))

# Sanity tests for these constraints.
dummy2_release := alt2
$(eval $(call mk_bin_rpm,dummy2,dummy))
$(eval \
	$(call not_overlaps,LessEpoch,dummy2) \
	$(call not_overlaps,EqualEpoch,dummy2) \
	$(call overlaps,GreaterEpoch,dummy2))
# TODO: dummy2Disttag
dummy0_release := alt0
$(eval $(call mk_bin_rpm,dummy0,dummy))
$(eval \
	$(call overlaps,LessEpoch,dummy0) \
	$(call not_overlaps,EqualEpoch,dummy0) \
	$(call not_overlaps,GreaterEpoch,dummy0))
dummyV2_version := 2
$(eval $(call mk_bin_rpm,dummyV2,dummy))
$(eval \
	$(call not_overlaps,LessEpoch,dummyV2) \
	$(call not_overlaps,EqualEpoch,dummyV2) \
	$(call overlaps,GreaterEpoch,dummyV2))
dummyV0_version := 0
$(eval $(call mk_bin_rpm,dummyV0,dummy))
$(eval \
	$(call overlaps,LessEpoch,dummyV0) \
	$(call not_overlaps,EqualEpoch,dummyV0) \
	$(call not_overlaps,GreaterEpoch,dummyV0))
dummyEpoch1_epoch := 1
$(eval $(call mk_bin_rpm,dummyEpoch1,dummy))
$(eval \
	$(call not_overlaps,LessEpoch,dummyEpoch1) \
	$(call overlaps,GreaterEpoch,dummyEpoch1) \
	$(call not_overlaps,EqualEpoch,dummyEpoch1)) # inexact match on Epoch

# Interesting tests with these constraints.
$(eval $(call not_overlaps,GreaterEpoch,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,LessEpoch,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call overlaps,EqualEpoch,dummy dummyDisttag))
# Only real names are obsoleted:
$(eval $(call overlaps_but_not_obsoletes,EqualEpoch,virtDummy virtDummyDisttag))


# Constraints without an epoch.
Greater = dummy > $(dummy_version)-$(dummy_release)
$(eval $(call mk_constraint_rpms,Greater))
Less = dummy < $(dummy_version)-$(dummy_release)
$(eval $(call mk_constraint_rpms,Less))
define Equal
dummy = $(dummy_version)-$(dummy_release)
endef
$(eval $(call mk_constraint_rpms,Equal))

# Sanity tests for these constraints.
$(eval \
	$(call not_overlaps,Less,dummy2) \
	$(call not_overlaps,Equal,dummy2) \
	$(call overlaps,Greater,dummy2))
$(eval \
	$(call overlaps,Less,dummy0) \
	$(call not_overlaps,Equal,dummy0) \
	$(call not_overlaps,Greater,dummy0))
$(eval \
	$(call not_overlaps,Less,dummyV2) \
	$(call not_overlaps,Equal,dummyV2) \
	$(call overlaps,Greater,dummyV2))
$(eval \
	$(call overlaps,Less,dummyV0) \
	$(call not_overlaps,Equal,dummyV0) \
	$(call not_overlaps,Greater,dummyV0))
$(eval \
	$(call not_overlaps,Less,dummyEpoch1) \
	$(call overlaps,Greater,dummyEpoch1,mishandled elusive dep with a release but no epoch) \
	$(call not_overlaps,Equal,dummyEpoch1,mishandled elusive dep with a release but no epoch)) # XFAIL (underspecified Epoch matches, but do we want this in this case?)

# Interesting tests with these constraints.
$(eval $(call not_overlaps,Greater,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,Less,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call overlaps,Equal,dummy dummyDisttag))
# Only real names are obsoleted:
$(eval $(call overlaps_but_not_obsoletes,Equal,virtDummy virtDummyDisttag))


# Constraints with an epoch and a disttag.
GreaterEpochDisttag = dummy > 0:$(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
$(eval $(call mk_constraint_rpms,GreaterEpochDisttag))
LessEpochDisttag = dummy < 0:$(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
$(eval $(call mk_constraint_rpms,LessEpochDisttag))
define EqualEpochDisttag
dummy = 0:$(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
endef
$(eval $(call mk_constraint_rpms,EqualEpochDisttag))


# Sanity tests for these constraints.
$(eval \
	$(call not_overlaps,LessEpochDisttag,dummy2) \
	$(call not_overlaps,EqualEpochDisttag,dummy2) \
	$(call overlaps,GreaterEpochDisttag,dummy2))
$(eval \
	$(call overlaps,LessEpochDisttag,dummy0) \
	$(call not_overlaps,EqualEpochDisttag,dummy0) \
	$(call not_overlaps,GreaterEpochDisttag,dummy0))
$(eval \
	$(call not_overlaps,LessEpochDisttag,dummyV2) \
	$(call not_overlaps,EqualEpochDisttag,dummyV2) \
	$(call overlaps,GreaterEpochDisttag,dummyV2))
$(eval \
	$(call overlaps,LessEpochDisttag,dummyV0) \
	$(call not_overlaps,EqualEpochDisttag,dummyV0) \
	$(call not_overlaps,GreaterEpochDisttag,dummyV0))
$(eval \
	$(call not_overlaps,LessEpochDisttag,dummyEpoch1) \
	$(call overlaps,GreaterEpochDisttag,dummyEpoch1) \
	$(call not_overlaps,EqualEpochDisttag,dummyEpoch1)) # inexact match on Epoch and Disttag

# Interesting tests with these constraints.
$(eval $(call not_overlaps,GreaterEpochDisttag,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,LessEpochDisttag,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,EqualEpochDisttag,dummy virtDummy)) # inexact match on Disttag
$(eval $(call overlaps,EqualEpochDisttag,dummyDisttag))
# Only real names are obsoleted:
$(eval $(call overlaps_but_not_obsoletes,EqualEpochDisttag,virtDummyDisttag))


# Constraints without an epoch and with a disttag.
GreaterDisttag = dummy > $(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
$(eval $(call mk_constraint_rpms,GreaterDisttag))
LessDisttag = dummy < $(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
$(eval $(call mk_constraint_rpms,LessDisttag))
define EqualDisttag
dummy = $(dummy_version)-$(dummy_release):$(dummyDisttag_disttag)
endef
$(eval $(call mk_constraint_rpms,EqualDisttag))


# Sanity tests for these constraints.
$(eval \
	$(call not_overlaps,LessDisttag,dummy2) \
	$(call not_overlaps,EqualDisttag,dummy2) \
	$(call overlaps,GreaterDisttag,dummy2))
$(eval \
	$(call overlaps,LessDisttag,dummy0) \
	$(call not_overlaps,EqualDisttag,dummy0) \
	$(call not_overlaps,GreaterDisttag,dummy0))
$(eval \
	$(call not_overlaps,LessDisttag,dummyV2) \
	$(call not_overlaps,EqualDisttag,dummyV2) \
	$(call overlaps,GreaterDisttag,dummyV2))
$(eval \
	$(call overlaps,LessDisttag,dummyV0) \
	$(call not_overlaps,EqualDisttag,dummyV0) \
	$(call not_overlaps,GreaterDisttag,dummyV0))
$(eval \
	$(call not_overlaps,LessDisttag,dummyEpoch1) \
	$(call overlaps,GreaterDisttag,dummyEpoch1,mishandled elusive dep with a release but no epoch)) # XFAIL
$(eval \
	$(call not_overlaps,EqualDisttag,dummyEpoch1)) # inexact match on Disttag
# TODO: to show the problem with elusive deps, match with dummyEpoch1 with disttag.

# Interesting tests with these constraints.
$(eval $(call not_overlaps,GreaterDisttag,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,LessDisttag,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,EqualDisttag,dummy virtDummy)) # inexact match on Disttag
$(eval $(call overlaps,EqualDisttag,dummyDisttag))
# Only real names are obsoleted:
$(eval $(call overlaps_but_not_obsoletes,EqualDisttag,virtDummyDisttag))


# Constraints with only version specified.
GreaterOnlyV = dummy > $(dummy_version)
$(eval $(call mk_constraint_rpms,GreaterOnlyV))
LessOnlyV = dummy < $(dummy_version)
$(eval $(call mk_constraint_rpms,LessOnlyV))
define EqualOnlyV
dummy = $(dummy_version)
endef
$(eval $(call mk_constraint_rpms,EqualOnlyV))

# Sanity tests for these constraints.
$(eval \
	$(call not_overlaps,LessOnlyV,dummy2) \
	$(call overlaps,EqualOnlyV,dummy2) \
	$(call not_overlaps,GreaterOnlyV,dummy2))
$(eval \
	$(call not_overlaps,LessOnlyV,dummy0) \
	$(call overlaps,EqualOnlyV,dummy0) \
	$(call not_overlaps,GreaterOnlyV,dummy0))
$(eval \
	$(call not_overlaps,LessOnlyV,dummyV2) \
	$(call not_overlaps,EqualOnlyV,dummyV2) \
	$(call overlaps,GreaterOnlyV,dummyV2))
$(eval \
	$(call overlaps,LessOnlyV,dummyV0) \
	$(call not_overlaps,EqualOnlyV,dummyV0) \
	$(call not_overlaps,GreaterOnlyV,dummyV0))
$(eval \
	$(call not_overlaps,LessOnlyV,dummyEpoch1) \
	$(call not_overlaps,GreaterOnlyV,dummyEpoch1) \
	$(call overlaps,EqualOnlyV,dummyEpoch1)) # underspecified Epoch matches

# Interesting tests with these constraints.
$(eval $(call not_overlaps,GreaterOnlyV,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call not_overlaps,LessOnlyV,dummy virtDummy dummyDisttag virtDummyDisttag))
$(eval $(call overlaps,EqualOnlyV,dummy dummyDisttag))
# Only real names are obsoleted:
$(eval $(call overlaps_but_not_obsoletes,EqualOnlyV,virtDummy virtDummyDisttag))
