# You can consider running these tests with
#
# HONOR_BUILDTIME=0
#
# which will result in --define '_upgrade_honor_buildtime 0';
# or with an empty string, then no option will be passed to rpm.
#
# or something else (but that's not very sensible, because I consider both cases
# where it is expected to be significant=important).

HONOR_BUILDTIME := 1
PRIORITY_DISTBRANCH := %nil

RPM_U = $(RPM) \
	$(and $(HONOR_BUILDTIME),--define '_upgrade_honor_buildtime $(HONOR_BUILDTIME)') \
	$(and $(PRIORITY_DISTBRANCH),--define '_priority_distbranch $(PRIORITY_DISTBRANCH)') \
	-U

############################################################
# Defs

# Upgradability must be a strict linear order ("newer"; below I denote it <).
# Only one thing must be true:
#
# a < b ("newer")
# a = b ("same new")
# a > b ("older")

# >= ("not newer")
define not_newer

nonupgradable_$(2)_after_$(1): $$$$($(1)) $$$$($(2))
	$$(RPM_U) $$($(1))
	! $$(RPM_U) $$($(2))

all_tests += $$(and $$($(1)),$$($(2)),nonupgradable_$(2)_after_$(1))
$(call give_xfail_reason,nonupgradable_$(2)_after_$(1),$(3))
endef

# basic helper
define upgradable

upgradable_$(2)_after_$(1): $$$$($(1)) $$$$($(2))
	$$(RPM_U) $$($(1))
	$$(RPM_U) $$($(2))
	[ `$$(filter-out --justdb,$$(RPM)) -qa | wc -l` -eq 1 ]

all_tests += $$(and $$($(1)),$$($(2)),upgradable_$(2)_after_$(1))
$(call give_xfail_reason,upgradable_$(2)_after_$(1),$(3))
endef

# = ("same new")
define same_new

$(call not_newer,$(1),$(2),$(3))
# ifneq ($(1),$(2))
$(if $(filter-out $(1),$(2)),$(call not_newer,$(2),$(1),$(3)))

# And this is a an even weaker statement:
# (Let it have a separate optional XFAIL reason.)
$(call same_new_without_honor_buildtime,$(1),$(2),$(4))
endef

# < ("newer")
define strictly_newer

$(call not_newer,$(2),$(1),$(3))
$(call upgradable,$(1),$(2),$(3))

$(call strictly_newer_without_honor_buildtime,$(1),$(2),$(4))
endef

define strictly_newer_without_honor_buildtime

$(call not_newer_without_honor_buildtime,$(2),$(1),$(3))
$(call upgradable,$(1)ALIAS,$(2)ALIAS,$(3))

upgradable_$(2)ALIAS_after_$(1)ALIAS: HONOR_BUILDTIME := 0
endef

define strictly_newer_iff_honor_buildtime

# it must be built later (according to the buildtime):
$$($(2)): sleep_after_$(1)

$(call not_newer,$(2),$(1),$(3))
$(call upgradable,$(1),$(2),$(3))

.PHONY: assert_$(2)_fresher_$(1)
assert_$(2)_fresher_$(1): $$$$($(1)) $$$$($(2))
	[ `$$(filter-out --justdb,$$(RPM)) -qp $$($(1)) --qf='%{BUILDTIME}'` \
	  -lt \
	  `$$(filter-out --justdb,$$(RPM)) -qp $$($(2)) --qf='%{BUILDTIME}'` \
	]

upgradable_$(2)_after_$(1) nonupgradable_$(1)_after_$(2): assert_$(2)_fresher_$(1)

# And now test the same packages without honoring buildtime.
# The xfail reason (if any) is the one passed separately as an additional
# parameter, because normally the tests without_honor_buildtime don't fail
# at the same time as the "main" tests.
$(call same_new_without_honor_buildtime,$(2),$(1),$(4))
endef

define not_newer_without_honor_buildtime

$(1)ALIAS = $$($(1))
$(2)ALIAS = $$($(2))
$(call not_newer,$(1)ALIAS,$(2)ALIAS,$(3))

nonupgradable_$(2)ALIAS_after_$(1)ALIAS: HONOR_BUILDTIME := 0
endef

define same_new_without_honor_buildtime

$(call not_newer_without_honor_buildtime,$(1),$(2),$(3))
# ifneq ($(1),$(2))
$(if $(filter-out $(1),$(2)),$(call not_newer_without_honor_buildtime,$(2),$(1),$(3)))
endef

############################################################
# Test cases

$(eval $(call same_new,dummy,dummy))
$(eval $(call same_new,dummyDisttag,dummyDisttag))


# We want another package with a more "fresh" release:
dummyNewRelease_release = $(dummy_release).1
$(eval $(call mk_bin_rpm,dummyNewRelease,dummy))
$(eval $(call same_new,dummyNewRelease,dummyNewRelease))

$(eval $(call strictly_newer,dummy,dummyNewRelease))

dummyDisttagNewRelease_release = $(dummyNewRelease_release)
$(eval $(call mk_bin_rpm,dummyDisttagNewRelease,dummyDisttag))
$(eval $(call same_new,dummyDisttagNewRelease,dummyDisttagNewRelease))

$(eval $(call strictly_newer,dummyDisttag,dummyNewRelease))
$(eval $(call strictly_newer,dummy,dummyDisttagNewRelease))
$(eval $(call strictly_newer,dummyDisttag,dummyDisttagNewRelease))


# We want another package with a less "fresh" release:
dummyOldRelease_release := 0
$(eval $(call mk_bin_rpm,dummyOldRelease,dummy))
# but with a more "fresh" buildtime
# (to exclude the effect of buildtime when the releases are significant):
$(dummyOldRelease): $(dummy)
$(eval $(call same_new,dummyOldRelease,dummyOldRelease))

$(eval $(call strictly_newer,dummyOldRelease,dummy))
$(eval $(call strictly_newer,dummyOldRelease,dummyNewRelease))

dummyDisttagOldRelease_release = $(dummyOldRelease_release)
$(eval $(call mk_bin_rpm,dummyDisttagOldRelease,dummyDisttag))
# but with a more "fresh" buildtime
# (to exclude the effect of buildtime when the releases are significant):
$(dummyDisttagOldRelease): $(dummyDisttag) $(dummy)
$(eval $(call same_new,dummyDisttagOldRelease,dummyDisttagOldRelease))

$(eval $(call strictly_newer,dummyOldRelease,dummyDisttag))
$(eval $(call strictly_newer,dummyDisttagOldRelease,dummy))
$(eval $(call strictly_newer,dummyDisttagOldRelease,dummyDisttag))
$(eval $(call strictly_newer,dummyOldRelease,dummyDisttagNewRelease))
$(eval $(call strictly_newer,dummyDisttagOldRelease,dummyNewRelease))
$(eval $(call strictly_newer,dummyDisttagOldRelease,dummyDisttagNewRelease))


# We want another package with the same release
# that will be built later (according to the buildtime):
$(eval $(call mk_bin_rpm,dummyFreshBuildtime,dummy))
$(eval $(call same_new,dummyFreshBuildtime,dummyFreshBuildtime))

$(eval $(call strictly_newer_iff_honor_buildtime,dummy,dummyFreshBuildtime))

# We want another package with the same release and disttag
# that will be built later (according to the buildtime):
$(eval $(call mk_bin_rpm,dummyDisttagFreshBuildtime,dummyDisttag))
$(eval $(call same_new,dummyDisttagFreshBuildtime,dummyDisttagFreshBuildtime))

$(eval $(call strictly_newer_iff_honor_buildtime,dummy,dummyDisttagFreshBuildtime))
$(eval $(call strictly_newer_iff_honor_buildtime,dummyDisttag,dummyFreshBuildtime))
$(eval $(call strictly_newer_iff_honor_buildtime,dummyDisttag,dummyDisttagFreshBuildtime))


# We want to test disttag comparison:
dummyDTagA2_disttag := aaa+222.2
$(eval $(call mk_bin_rpm,dummyDTagA2,dummy))
$(eval $(call same_new,dummyDTagA2,dummyDTagA2))
dummyDTagB1_disttag := bbb+111.1
$(eval $(call mk_bin_rpm,dummyDTagB1,dummy))
$(eval $(call same_new,dummyDTagB1,dummyDTagB1))
$(eval $(call mk_bin_rpm,dummyDTagA2FreshBuildtime,dummyDTagA2))
$(eval $(call same_new,dummyDTagA2FreshBuildtime,dummyDTagA2FreshBuildtime))
# Force the buildtime order:
$(dummyDTagB1): sleep_after_dummyDTagA2
$(dummyDTagA2FreshBuildtime): sleep_after_dummyDTagB1

$(eval $(call strictly_newer,dummyDTagA2,dummyDTagB1,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL
$(eval $(call strictly_newer,dummyDTagA2FreshBuildtime,dummyDTagB1,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL


# Test upgrade to another package with the same disttag
# that is built later (according to the buildtime):
$(eval $(call strictly_newer_iff_honor_buildtime,dummyDTagA2,dummyDTagA2FreshBuildtime))


# We want another package with almost the same disttag
# that will be built later (according to the buildtime):
dummyDTagA1FreshBuildtime_disttag := aaa+111.11
$(eval $(call mk_bin_rpm,dummyDTagA1FreshBuildtime,dummyDTagA2))
$(eval $(call same_new,dummyDTagA1FreshBuildtime,dummyDTagA1FreshBuildtime))

$(eval $(call strictly_newer_iff_honor_buildtime,dummyDTagA2,dummyDTagA1FreshBuildtime))


# We want to test disttag comparison
# and include the obsolete (dot) and current format (plus).
# We repeat the same tests as above (DTagA -> DTagPlus, DTagB -> DTagDot);
# dummyDTagB1 is replaced with a pkg with an obsolete disttag format
# (and must be older unlike above, because that's a more interesting case).
dummyDTagPlus2_disttag := y7.1+222.2
$(eval $(call mk_bin_rpm,dummyDTagPlus2,dummy))
$(eval $(call same_new,dummyDTagPlus2,dummyDTagPlus2))
dummyDTagDot3_disttag := y7.1.333.3
$(eval $(call mk_bin_rpm,dummyDTagDot3,dummy))
$(eval $(call same_new,dummyDTagDot3,dummyDTagDot3))
$(eval $(call mk_bin_rpm,dummyDTagPlus2FreshBuildtime,dummyDTagPlus2))
$(eval $(call same_new,dummyDTagPlus2FreshBuildtime,dummyDTagPlus2FreshBuildtime))
# Force the buildtime order:
$(dummyDTagDot3): sleep_after_dummyDTagPlus2
$(dummyDTagPlus2FreshBuildtime): sleep_after_dummyDTagDot3

$(eval $(call strictly_newer,dummyDTagDot3,dummyDTagPlus2,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL
$(eval $(call strictly_newer,dummyDTagDot3,dummyDTagPlus2FreshBuildtime,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL


# Again, We want to test disttag comparison
# and include the obsolete (dot) and current format (plus).
# But now the current disttag's branch is %_priority_distbranch.
# We call those packages DTagPri instead of DTagPlus.
#
# We make a dirty global override of the neutral value of PRIORITY_DISTBRANCH
# (it must not affect other tests, because we don't use such disttags elsewhere).
# TODO: rule-local override.
PRIORITY_DISTBRANCH := x7.1
dummyDTagPri2_disttag := x7.1+222.2
$(eval $(call mk_bin_rpm,dummyDTagPri2,dummy))
$(eval $(call same_new,dummyDTagPri2,dummyDTagPri2))
$(eval $(call mk_bin_rpm,dummyDTagPri2FreshBuildtime,dummyDTagPri2))
$(eval $(call same_new,dummyDTagPri2FreshBuildtime,dummyDTagPri2FreshBuildtime))
# Force the buildtime order:
$(dummyDTagDot3): sleep_after_dummyDTagPri2
$(dummyDTagPri2FreshBuildtime): sleep_after_dummyDTagDot3

$(eval $(call strictly_newer,dummyDTagDot3,dummyDTagPri2,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL
$(eval $(call strictly_newer,dummyDTagDot3,dummyDTagPri2FreshBuildtime,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL

# Also test that if both disttags are from the PRIORITY_DISTBRANCH,
# the general rules apply.
# (The tests are the same as above for dummyDTagA?{,FreshBuildtime}.)

# Test upgrade to another package with the same disttag (from PRIORITY_DISTBRANCH)
# that is built later (according to the buildtime):
$(eval $(call strictly_newer_iff_honor_buildtime,dummyDTagPri2,dummyDTagPri2FreshBuildtime))


# We want another package with almost the same disttag (from PRIORITY_DISTBRANCH)
# that will be built later (according to the buildtime):
dummyDTagPri1FreshBuildtime_disttag := x7.1+111.11
$(eval $(call mk_bin_rpm,dummyDTagPri1FreshBuildtime,dummyDTagPri2))
$(eval $(call same_new,dummyDTagPri1FreshBuildtime,dummyDTagPri1FreshBuildtime))

$(eval $(call strictly_newer_iff_honor_buildtime,dummyDTagPri2,dummyDTagPri1FreshBuildtime))


# Finally, test the %_priority_distbranch feature in a normal situation.
# Force the buildtime order:
$(dummyDTagPlus2): sleep_after_dummyDTagPri2
$(dummyDTagPri2FreshBuildtime): sleep_after_dummyDTagPlus2
$(eval $(call strictly_newer,dummyDTagPlus2,dummyDTagPri2,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL
$(eval $(call strictly_newer,dummyDTagPlus2,dummyDTagPri2FreshBuildtime,\
	,\
	fails in non-standard configuration without honor_buildtime)) # XFAIL
